from collections import defaultdict
from featuretools import variable_types
from featuretools.variable_types import Categorical, Numeric, Boolean, Ordinal
from featuretools.primitives.api import (IdentityFeature, BinaryFeature,
                                       TimeSince, DirectFeature, Compare,
                                       AggregationFeature)
import featuretools.primitives.api as ftypes
from .dfs_filters import (TraverseUp, AggFeatStack, LimitModeUniques,
                          WhereFeatsDifferentThanBaseFeats,
                          LimitWhereClausesByFeatureType,
                          LimitWhereClausesByVariableType,
                          DFeatWhere, WhereStack)

import logging
logger = logging.getLogger('featuretools')


class DeepFeatureSynthesis(object):
    """Automatically produce features for a target entity in an Entityset.

        Args:
            target_entity_id (str): id of entity to build features for

            entityset (:class:`.EntitySet`): Entityset to build features for

            filters (list[:class:`.DFSFilterBase`], optional) : list of dfs filters
                to apply.

                Default:

                [:class:`synthesis.dfs_filters.TraverseUp`,
                 :class:`synthesis.dfs_filters.LimitModeUniques`,
                 :class:`synthesis.dfs_filters.LimitWhereClausesByVariableType`,
                 :class:`synthesis.dfs_filters.LimitWhereClausesByFeatureType`,
                 :class:`synthesis.dfs_filters.WhereFeatsDifferentThanBaseFeats`,
                 :class:`synthesis.dfs_filters.WhereStack`,
                 :class:`synthesis.dfs_filters.DFeatWhere`,
                 :class:`synthesis.dfs_filters.AggFeatStack`]

            agg_primitives (list[:class:`features.AggregationFeature`], optional):
                list of Aggregation Feature types to apply.

                Default:

                [:class:`features.Sum`, :class:`features.Std`,
                 :class:`features.Max`, :class:`features.Skew`,
                 :class:`features.Min`, :class:`features.Mean`,
                 :class:`features.Count`, :class:`features.PercentTrue`,
                 :class:`features.NUnique`, :class:`features.Mode`]

            trans_primitives (list[:class:`features.TransformFeature`], optional):
                list of Transform Feature functions to apply.

                Default:

                [:class:`features.Day`, :class:`features.Year`,
                 :class:`features.Month`, :class:`features.Weekday`]

            where_primitives (list[:class:`features.AggregationFeature`], optional):
                only add where clauses to these types of Aggregation
                Features.

                Default:

                    [:class:`features.Count`]

            max_depth (int, optional) : maximum allowed depth of features.
                Default: 2. If -1, no limit.

            max_hlevel (int, optional) :  #TODO how to document.
                Default: 2. If -1, no limit.

            max_features (int, optional) : Cap the number of generated features to
                this number. If -1, no limit.

            allowed_paths (list[list[str]], optional): Allowed entity paths to make
                features for. If None, use all paths.

            ignore_entities (list[str], optional): List of entities to
                blacklist when creating features. If None, use all entities

            ignore_variables (dict[str : str], optional): List of specific
                variables within each entity to blacklist when creating features.
                If None, use all variables

            seed_features (list[:class:`.FeatureBase`], optional): List of manually
                defined features to use.

            drop_contains (list[str], optional): drop features
                that contains these strings in name

            drop_exact (list[str], optional): drop features that
                exactly match these strings in name
        """

    def __init__(self,
                 target_entity_id,
                 entityset,
                 filters=None,
                 agg_primitives=None,
                 trans_primitives=None,
                 where_primitives=None,
                 max_depth=None,
                 max_hlevel=None,
                 max_features=None,
                 allowed_paths=None,
                 ignore_entities=None,
                 ignore_variables=None,
                 seed_features=None,
                 drop_contains=None,
                 drop_exact=None):

        if max_depth is None:
            max_depth = 2
        elif max_depth == -1:
            max_depth = None
        self.max_depth = max_depth

        if max_hlevel is None:
            max_hlevel = 2
        elif max_hlevel == -1:
            max_hlevel = None
        self.max_hlevel = max_hlevel

        if max_features is None:
            max_features = -1
        self.max_features = max_features

        self.allowed_paths = allowed_paths
        if self.allowed_paths:
            self.allowed_paths = set()
            for path in allowed_paths:
                self.allowed_paths.add(tuple(path))

        if ignore_entities is None:
            self.ignore_entities = set()
        else:
            assert target_entity_id not in ignore_entities,\
                "Can't ignore target_entity!"
            self.ignore_entities = set(ignore_entities)

        self.ignore_variables = defaultdict(set)
        if ignore_variables is not None:
            for eid, vars in ignore_variables.iteritems():
                self.ignore_variables[eid] = set(vars)

        self.target_entity_id = target_entity_id
        self.es = entityset

        if where_primitives is None:
            where_primitives = [ftypes.Count]

        if filters is None:
            filters = [TraverseUp(),
                       LimitModeUniques(),
                       LimitWhereClausesByVariableType([Categorical]),
                       LimitWhereClausesByFeatureType(where_primitives),
                       WhereFeatsDifferentThanBaseFeats(),
                       WhereStack(),
                       DFeatWhere(),
                       AggFeatStack()]

        self.pre_instance_filters = []
        self.post_instance_filters = []
        self.traversal_filters = []

        for f in filters:
            if f.filter_type == 'pre_instance':
                self.pre_instance_filters.append(f)
            elif f.filter_type == 'post_instance':
                self.post_instance_filters.append(f)
            elif f.filter_type == 'traversal':
                self.traversal_filters.append(f)
            else:
                raise NotImplementedError("Unknown filter type {}"
                                          .format(f.filter_type))

        if agg_primitives is None:
            agg_primitives = [ftypes.Sum, ftypes.Std, ftypes.Max, ftypes.Skew,
                              ftypes.Min, ftypes.Mean, ftypes.Count,
                              ftypes.PercentTrue, ftypes.NUnique, ftypes.Mode]
        self.agg_primitives = agg_primitives

        if trans_primitives is None:
            trans_primitives = [ftypes.Day, ftypes.Year, ftypes.Month, ftypes.Weekday]  # ftypes.TimeSince
        self.trans_primitives = trans_primitives

        self.seed_features = seed_features or []
        self.drop_exact = drop_exact or []
        self.drop_contains = drop_contains or []

    def build_features(self, variable_types=None):
        """Automatically builds feature definitions for target
            entity using Deep Feature Synthesis algorithm

        Args:
            variable_types (list[:class:`variable_types.Variable`] or str,
                optional): Types of variables to return. If None, default to
                Numeric, Categorical, Ordinal, and Boolean. If given as
                the string 'all', use all available variable types.

        Returns:
            list[:class:`features.BaseFeature`]: returns a list of
                features for target entity, sorted by feature depth
                (shallow first)
        """
        all_features = {}
        for e in self.es.entities:
            if e not in self.ignore_entities:
                all_features[e.id] = {}

        # add seed features, if any, for dfs to build on top of
        if self.seed_features is not None:
            for f in self.seed_features:
                self._handle_new_feature(all_features=all_features,
                                         new_feature=f)

        self.where_clauses = defaultdict(set)
        self._run_dfs(self.es[self.target_entity_id], [],
                      all_features, max_depth=self.max_depth)

        new_features = all_features[self.target_entity_id].values()

        if variable_types is None:
            variable_types = [Numeric,
                              Categorical,
                              Ordinal,
                              Boolean]
        elif variable_types == 'all':
            variable_types = None
        else:
            msg = "variable_types must be a list, or 'all'"
            assert isinstance(variable_types, list), msg

        if variable_types is not None:
            new_features = [f for f in new_features
                            if f.variable_type in variable_types]

        def check_secondary_index(f):
            secondary_time_index = self.es[self.target_entity_id].secondary_time_index
            for s_time_index, exclude in secondary_time_index.items():
                if isinstance(f, IdentityFeature) and f.variable.id in exclude:
                    return False
                elif isinstance(f, (BinaryFeature, Compare)):
                    if (not check_secondary_index(f.left) or
                            not check_secondary_index(f.right)):
                        return False
                if isinstance(f, TimeSince) and not check_secondary_index(f.base_features[0]):
                    return False

            return True

        def filt(f):
            # remove identity features of the ID field of the target entity
            if (isinstance(f, IdentityFeature) and
                    f.entity.id == self.target_entity_id and
                    f.variable.id == self.es[self.target_entity_id].id_variable):
                return False

            if (isinstance(f, (IdentityFeature, BinaryFeature, Compare, TimeSince)) and
                    not check_secondary_index(f)):

                return False

            return True

        new_features = filter(filt, new_features)

        # sanity check for duplicate features
        l = [f.hash() for f in new_features]
        assert len(set([f for f in l if l.count(f) > 1])) == 0, \
            'Multiple features with same name' + \
            str(set([f for f in l if l.count(f) > 1]))

        new_features.sort(key=lambda f: f._get_depth())

        new_features = self._filter_features(new_features)

        if self.max_features > 0:
            new_features = new_features[:self.max_features]

        return new_features

    def _filter_features(self, features):
        assert isinstance(self.drop_exact, list), "drop_exact must be a list"
        assert isinstance(self.drop_contains, list), "drop_contains must be a list"

        f_keep = []
        for f in features:
            keep = True
            for contains in self.drop_contains:
                if contains in f.get_name():
                    keep = False
                    break

            if f.get_name() in self.drop_exact:
                keep = False

            if keep:
                f_keep.append(f)

        return f_keep

    def _run_dfs(self, entity, entity_path, all_features, max_depth):
        """
        create features for the provided entity

        Args:
            entity (:class:`.Entity`): entity to create features for
            entity_path (list[str]): list of entity ids
            all_features (dict[:class:`.Entity`.id:dict->[str->:class:`BaseFeature`]]):
                Dict containing a dict for each entity. Each nested dict
                has features as values with their ids as keys
            max_depth (int) : maximum allowed depth of features
        """
        if max_depth is not None and max_depth < 0:
            return

        entity_path.append(entity.id)
        """
        Step 1 - Recursively build features for each entity in a backward relationship
        """

        backward_entities = self.es.get_backward_entities(entity.id)
        backward_entities = [b_id for b_id in backward_entities
                             if self._apply_traversal_filters(entity, self.es[b_id],
                                                              entity_path,
                                                              forward=False) and
                             b_id not in self.ignore_entities]
        for b_entity_id in backward_entities:
            # if in path, we've alrady built features
            if b_entity_id in entity_path:
                continue

            if self.allowed_paths and tuple(entity_path + [b_entity_id]) not in self.allowed_paths:
                continue
            new_max_depth = None
            if max_depth is not None:
                new_max_depth = max_depth - 1
            self._run_dfs(entity=self.es[b_entity_id],
                          entity_path=list(entity_path),
                          all_features=all_features,
                          max_depth=new_max_depth)

        """
        Step 2 - Create agg_feat features for all deep backward relationships
        """

        # search for deep relationships of that passed the traversal filter
        build_entities = list(backward_entities)
        for b_id in backward_entities:
            build_entities += self.es.get_backward_entities(b_id, deep=True)

        for b_entity_id in build_entities:
            if b_entity_id in self.ignore_entities:
                continue
            if self.allowed_paths and tuple(entity_path + [b_entity_id]) not in self.allowed_paths:
                continue
            self._build_backward_features(parent_entity=self.es[entity.id],
                                          child_entity=self.es[b_entity_id],
                                          all_features=all_features,
                                          max_depth=max_depth)

        """
        Step 3 - Create Transform features
        """
        self._build_transform_features(all_features, entity, max_depth=max_depth)

        """
        Step 4 - Recursively build features for each entity in a forward relationship
        """
        forward_entities = self.es.get_forward_entities(entity.id)
        # filter entities in path and using traversal filters
        forward_entities = [f_id for f_id in forward_entities
                            if self._apply_traversal_filters(entity,
                                                             self.es[f_id],
                                                             entity_path) and
                            f_id not in self.ignore_entities]

        for f_entity_id in forward_entities:
            # if in path, we've already built features
            if f_entity_id in entity_path:
                continue

            if self.allowed_paths and tuple(entity_path + [f_entity_id]) not in self.allowed_paths:
                continue

            new_max_depth = None
            if max_depth is not None:
                new_max_depth = max_depth - 1
            self._run_dfs(entity=self.es[f_entity_id],
                          entity_path=list(entity_path),
                          all_features=all_features,
                          max_depth=new_max_depth)

        """
        Step 5 - Create dfeat features for forward relationships
        """
        # get forward relationship involving forward entities
        forward = [r for r in self.es.get_forward_relationships(entity.id)
                   if r.parent_entity.id in forward_entities and
                   r.parent_entity.id not in self.ignore_entities]

        for r in forward:
            if self.allowed_paths and tuple(entity_path + [r.parent_entity.id]) not in self.allowed_paths:
                continue
            self._build_forward_features(all_features=all_features,
                                         parent_entity=r.parent_entity,
                                         child_entity=r.child_entity,
                                         relationship=r,
                                         max_depth=max_depth)

    def _apply_traversal_filters(self, parent_entity, child_entity,
                                 entity_path, forward=True):
        for f in self.traversal_filters:
            if not f.is_valid(entity=parent_entity,
                              child_entity=child_entity,
                              target_entity_id=self.target_entity_id,
                              entity_path=entity_path, forward=forward):
                return False

        return True

    def _handle_new_feature(self, new_feature, all_features):
        """Adds new feature to the dict

        Args:
            new_feature (:class:`.FeatureBase`): New feature being
                checked
            all_features (dict[:class:`.Entity`.id:dict->[str->:class:`BaseFeature`]]):
                Dict containing a dict for each entity. Each nested dict
                has features as values with their ids as keys

        Returns:
            dict{:class:`.FeatureBase`: {featureid: feature}: Dict of
                features with any new features

        Raises:
            Exception: Attempted to add a single feature multiple times
        """
        # check if new_feature is already added: "in" uses the __eq__ function
        # so this will work.

        if (self.max_hlevel is not None and
                self._max_hlevel(new_feature) > self.max_hlevel):
            return
        entity_id = new_feature.entity.id
        if new_feature.hash() in all_features[entity_id]:
            return
            raise Exception("DFS runtime error: tried to add feature %s"
                            " more than once" % (new_feature.get_name()))

        # update the dict
        all_features[entity_id][new_feature.hash()] = new_feature

    def _add_identity_features(self, all_features, entity):
        """converts all variables from the given entity into features

        Args:
            all_features (dict[:class:`.Entity`.id:dict->[str->:class:`BaseFeature`]]):
                Dict containing a dict for each entity. Each nested dict
                has features as values with their ids as keys
            entity (:class:`.Entity`): entity to calculate features for
        """
        variables = entity.variables
        ignore_variables = self.ignore_variables[entity.id]
        for v in variables:
            if v.id in ignore_variables:
                continue
            new_f = IdentityFeature(variable=v, name=v.name)
            self._handle_new_feature(all_features=all_features,
                                     new_feature=new_f)

    def _build_where_clauses(self, all_features, entity):
        """Traverses all identity features and creates a Compare for
            each one, based on some heuristics

        Args:
            all_features (dict[:class:`.Entity`.id:dict->[str->:class:`BaseFeature`]]):
                Dict containing a dict for each entity. Each nested dict
                has features as values with their ids as keys
            entity (:class:`.Entity`): entity to calculate features for
        """
        identities = [f for _, f in all_features[entity.id].iteritems()
                      if isinstance(f, IdentityFeature)]

        for feat in identities:
            if feat.variable.interesting_values is None:
                continue

            for val in feat.variable.interesting_values:
                self.where_clauses[entity.id].add(Compare(feat, Compare.EQ, val))

    def _build_transform_features(self, all_features, entity, max_depth=0):
        """Creates trans_features for all the variables in an entity

        Args:
            all_features (dict[:class:`.Entity`.id:dict->[str->:class:`BaseFeature`]]):
                Dict containing a dict for each entity. Each nested dict
                has features as values with their ids as keys
            entity (:class:`.Entity`): entity to calculate features for
        """
        if max_depth is not None and max_depth < 0:
            return

        self._add_identity_features(all_features, entity)

        features = self._features_by_type(all_features, entity,
                                          variable_types.PandasTypes._all,
                                          max_depth)


        relation_vars = []
        for r in entity.entityset.get_forward_relationships(entity.id):
            relation_vars.append(r.child_variable)

        # TODO:
        # should we include if not in path to target entity?



        to_add =[]
        for trans_prim in self.trans_primitives:       # list of builder classes
            if entity.has_time_index() and trans_prim.entity_groupby and trans_prim.require_time_index:
                for rv in relation_vars:
                    new_f = trans_prim(entity[entity.time_index],  group_feature=rv)
                    if new_f.can_apply(entity=entity):
                        self._handle_new_feature(all_features=all_features,
                                                 new_feature=new_f)


            for f in features:
                if trans_prim.entity_groupby:
                    # this handles the TimBetwe
                    if trans_prim.require_time_index:
                        continue
                    for rv in relation_vars:
                        try:
                            new_f = trans_prim(f, group_feature=rv)
                        except:
                            pass
                        if new_f.can_apply(entity=entity):
                            self._handle_new_feature(all_features=all_features,
                                                     new_feature=new_f)
                else:
                    try:
                        new_f = trans_prim(f)
                        if new_f.can_apply(entity=entity):
                            self._handle_new_feature(all_features=all_features,
                                                     new_feature=new_f)
                    except:
                        pass


        # now that all transform features are added, build where clauses
        self._build_where_clauses(all_features, entity)

    def _build_forward_features(self, all_features, parent_entity,
                                child_entity, relationship, max_depth=0):

        if max_depth is not None and max_depth < 0:
            return

        features = self._features_by_type(all_features=all_features,
                                          entity=parent_entity,
                                          variable_type=[Numeric,
                                                         Categorical,
                                                         Ordinal],
                                          max_depth=max_depth)

        for f in features:
            if self._feature_in_relationship_path([relationship], f):
                continue

            new_f = DirectFeature(base_feature=f,
                                  child_entity=child_entity)
            self._handle_new_feature(all_features=all_features,
                                     new_feature=new_f)

    def _build_backward_features(self, all_features,
                                 parent_entity, child_entity, max_depth=0):
        if max_depth is not None and max_depth < 0:
            return
        relationship_path = self.es.find_backward_path(parent_entity.id,
                                                       child_entity.id)

        new_max_depth = None
        if max_depth is not None:
            new_max_depth = max_depth - 1
        features = self._features_by_type(all_features=all_features,
                                          entity=child_entity,
                                          variable_type=[Numeric,
                                                         Boolean,
                                                         Categorical,
                                                         Ordinal],
                                          max_depth=new_max_depth)

        for agg_feat in self.agg_primitives:
            for f in features:
                if self._feature_in_relationship_path(relationship_path, f):
                    continue

                wheres = [None] + list(self.where_clauses[child_entity.id])
                for where in wheres:
                    if not self._check_pre_instance_filters(agg_feat, f,
                                                            child_entity,
                                                            where=where):
                        continue

                    try:
                        new_f = agg_feat(base_feature=f,
                                         parent_entity=parent_entity,
                                         where=where)
                    except:
                        continue
                    else:
                        if new_f.can_apply(parent_entity,
                                           self.target_entity_id,
                                           filters=self.post_instance_filters):

                            self._handle_new_feature(new_f, all_features)

    def _check_pre_instance_filters(self, feature, child_feature,
                                    child_entity, where=None):

        for filt in self.pre_instance_filters:
            if not filt.is_valid(feature=feature, child_feature=child_feature,
                                 child_entity=child_entity, where=where):
                return False
        return True

    def _features_by_type(self, all_features, entity, variable_type, max_depth):
        selected_features = []
        if max_depth is not None and max_depth < 0:
            return selected_features

        for feat in all_features[entity.id]:
            f = all_features[entity.id][feat]

            if (variable_type == variable_types.PandasTypes._all or
                    f.variable_type == variable_type or
                    f.variable_type in variable_type):
                if ((max_depth is None or self._feature_depth(f) <= max_depth) and
                        (self.max_hlevel is None or
                         self._max_hlevel(f) <= self.max_hlevel)):
                    selected_features.append(f)
                selected_features.append(f)
        return selected_features

    def _feature_in_relationship_path(self, relationship_path, feature):
        # must be identity feature to be in the relationship path
        if not isinstance(feature, IdentityFeature):
            return False

        for relationship in relationship_path:
            if relationship.child_entity == feature.entity and \
               relationship.child_variable == feature.variable:
                return True

            if relationship.parent_entity == feature.entity and \
               relationship.parent_variable == feature.variable:
                return True

        return False

    def _feature_depth(self, f):
        return f._get_depth()

    def _max_hlevel(self, f):
        # for each base_feat along each path in f,
        # if base_feat is a direct_feature of an agg_feature
        # determine aggfeat's hlevel
        # return max hlevel
        deps = [f] + f._get_deep_dependencies()
        hlevel = 0
        for d in deps:
            if isinstance(d, DirectFeature) and \
                    isinstance(d.base_feature, AggregationFeature):

                assert d.parent_entity.id == d.base_feature.entity.id
                path, new_hlevel = self.es.find_path(self.target_entity_id,
                                                     d.parent_entity.id,
                                                     include_num_forward=True)
                hlevel = max(hlevel, new_hlevel)
        return hlevel
